@@ -37,14 +37,15 @@
 #define SQLITE_INDEX_CONSTRAINT_ISNULL    71
 #endif
 
+/* Properties are additional columns after graph and rowid */
+#define FIRST_PROPERTY_COLUMN 2
+
 enum {
-	COL_ROWID,
 	COL_GRAPH,
 	COL_SUBJECT,
 	COL_PREDICATE,
 	COL_OBJECT,
 	COL_OBJECT_TYPE,
-	COL_OBJECT_RAW,
 	N_COLS
 };
 
@@ -78,6 +79,8 @@ typedef struct {
 	struct sqlite3_vtab_cursor parent;
 	TrackerTriplesVTab *vtab;
 	struct sqlite3_stmt *stmt;
+	TrackerProperty **column_properties;
+	guint n_alloc_column_properties;
 
 	struct {
 		sqlite3_value *graph;
@@ -89,8 +92,14 @@ typedef struct {
 
 	GHashTable *query_graphs;
 	GList *properties;
+	GList *classes;
 	GList *graphs;
 
+	const GList *cur_property;
+	const GList *cur_class;
+	const GList *cur_graph;
+	gint column;
+
 	guint64 rowid;
 	guint finished : 1;
 } TrackerTriplesCursor;
@@ -121,6 +130,7 @@ tracker_triples_cursor_reset (TrackerTriplesCursor *cursor)
 	g_clear_pointer (&cursor->match.subject, sqlite3_value_free);
 	g_clear_pointer (&cursor->match.predicate, sqlite3_value_free);
 	g_clear_pointer (&cursor->properties, g_list_free);
+	g_clear_pointer (&cursor->classes, g_list_free);
 	g_clear_pointer (&cursor->graphs, g_list_free);
 	g_clear_pointer (&cursor->query_graphs, g_hash_table_unref);
 	cursor->match.idxFlags = 0;
@@ -134,6 +144,7 @@ tracker_triples_cursor_free (gpointer data)
 	TrackerTriplesCursor *cursor = data;
 
 	tracker_triples_cursor_reset (cursor);
+	g_clear_pointer (&cursor->column_properties, g_free);
 	g_free (cursor);
 }
 
@@ -154,13 +165,11 @@ triples_connect (sqlite3            *db,
 
 	rc = sqlite3_declare_vtab (module->db,
 	                           "CREATE TABLE x("
-	                           "    ID INTEGER,"
 	                           "    graph INTEGER,"
 	                           "    subject INTEGER, "
 	                           "    predicate INTEGER, "
 	                           "    object INTEGER, "
-	                           "    object_type INTEGER, "
-	                           "    object_raw INTEGER"
+	                           "    object_type INTEGER "
 	                           ")");
 
 	if (rc == SQLITE_OK) {
@@ -202,15 +211,9 @@ triples_best_index (sqlite3_vtab       *vtab,
 		 * translation to strings can be done.
 		 */
 		if (info->aConstraint[i].iColumn == COL_OBJECT ||
-		    info->aConstraint[i].iColumn == COL_OBJECT_TYPE ||
-		    info->aConstraint[i].iColumn == COL_OBJECT_RAW)
+		    info->aConstraint[i].iColumn == COL_OBJECT_TYPE)
 			continue;
 
-		if (info->aConstraint[i].iColumn == COL_ROWID) {
-			sqlite3_free (idx_str);
-			return SQLITE_ERROR;
-		}
-
 		/* We can only check for (in)equality */
 		if (info->aConstraint[i].op != SQLITE_INDEX_CONSTRAINT_EQ &&
 		    info->aConstraint[i].op != SQLITE_INDEX_CONSTRAINT_NE &&
@@ -221,11 +224,11 @@ triples_best_index (sqlite3_vtab       *vtab,
 		}
 
 		/* idxNum encodes the used columns and their operators */
-		idx |= masks[info->aConstraint[i].iColumn - 1].mask;
+		idx |= masks[info->aConstraint[i].iColumn].mask;
 
 		if (info->aConstraint[i].op == SQLITE_INDEX_CONSTRAINT_NE ||
 		    info->aConstraint[i].op == SQLITE_INDEX_CONSTRAINT_ISNOTNULL)
-			idx |= masks[info->aConstraint[i].iColumn - 1].negated_mask;
+			idx |= masks[info->aConstraint[i].iColumn].negated_mask;
 
 		/* idxStr stores the mapping between columns and filter arguments */
 		idx_str[info->aConstraint[i].iColumn] = argv_idx - 1;
@@ -284,26 +287,48 @@ triples_close (sqlite3_vtab_cursor *vtab_cursor)
 }
 
 static void
-collect_properties (TrackerTriplesCursor *cursor)
+collect_tables (TrackerTriplesCursor *cursor)
 {
-	TrackerProperty **properties;
-	guint n_properties, i;
+	TrackerProperty *property = NULL;
+	const gchar *uri = NULL;
+	gboolean pred_negated;
 
-	properties = tracker_ontologies_get_properties (cursor->vtab->module->ontologies,
-							&n_properties);
-	for (i = 0; i < n_properties; i++) {
-		if (cursor->match.predicate) {
-			gboolean negated = !!(cursor->match.idxFlags & IDX_MATCH_PREDICATE_NEG);
-			gboolean equals =
-				(sqlite3_value_int64 (cursor->match.predicate) ==
-				 tracker_property_get_id (properties[i]));
+	pred_negated = !!(cursor->match.idxFlags & IDX_MATCH_PREDICATE_NEG);
 
-			if (equals == negated)
-				continue;
+	if (cursor->match.predicate) {
+		uri = tracker_ontologies_get_uri_by_id (cursor->vtab->module->ontologies,
+		                                        sqlite3_value_int64 (cursor->match.predicate));
+	}
+
+	if (uri) {
+		property = tracker_ontologies_get_property_by_uri (cursor->vtab->module->ontologies,
+		                                                   uri);
+	}
+
+	if (property && !pred_negated) {
+		cursor->properties = g_list_prepend (cursor->properties, property);
+	} else {
+		TrackerProperty **properties;
+		guint n_properties, i;
+
+		properties = tracker_ontologies_get_properties (cursor->vtab->module->ontologies,
+		                                                &n_properties);
+		for (i = 0; i < n_properties; i++) {
+			if (tracker_property_get_multiple_values (properties[i])) {
+				if (pred_negated && property == properties[i])
+					continue;
+
+				cursor->properties = g_list_prepend (cursor->properties,
+				                                     properties[i]);
+			} else {
+				TrackerClass *class;
+
+				class = tracker_property_get_domain (properties[i]);
+
+				if (!g_list_find (cursor->classes, class))
+					cursor->classes = g_list_prepend (cursor->classes, class);
+			}
 		}
-
-		cursor->properties = g_list_prepend (cursor->properties,
-		                                     properties[i]);
 	}
 }
 
@@ -356,40 +381,6 @@ collect_graphs (TrackerTriplesCursor *cursor)
 	return rc;
 }
 
-static gchar *
-convert_to_string (const gchar         *table_name,
-		   TrackerPropertyType  type)
-{
-	switch (type) {
-	case TRACKER_PROPERTY_TYPE_STRING:
-	case TRACKER_PROPERTY_TYPE_LANGSTRING:
-	case TRACKER_PROPERTY_TYPE_INTEGER:
-		return g_strdup_printf ("t.\"%s\"", table_name);
-	case TRACKER_PROPERTY_TYPE_RESOURCE:
-		return g_strdup_printf ("(SELECT COALESCE(Uri, 'urn:bnode:' || ID) FROM Resource WHERE ID = t.\"%s\")",
-		                        table_name);
-	case TRACKER_PROPERTY_TYPE_BOOLEAN:
-		return g_strdup_printf ("CASE t.\"%s\" "
-		                        "WHEN 1 THEN 'true' "
-		                        "WHEN 0 THEN 'false' "
-		                        "ELSE NULL END",
-		                        table_name);
-	case TRACKER_PROPERTY_TYPE_DATE:
-		return g_strdup_printf ("strftime (\"%%Y-%%m-%%d\", t.\"%s\", \"unixepoch\")",
-		                        table_name);
-	case TRACKER_PROPERTY_TYPE_DATETIME:
-		return g_strdup_printf ("SparqlFormatTime (t.\"%s\")",
-		                        table_name);
-        case TRACKER_PROPERTY_TYPE_DOUBLE:
-	case TRACKER_PROPERTY_TYPE_UNKNOWN:
-		/* Let sqlite convert the expression to string */
-		return g_strdup_printf ("CAST (t.\"%s\" AS TEXT)",
-		                        table_name);
-	}
-
-	g_assert_not_reached();
-}
-
 static void
 add_arg_check (GString       *str,
 	       sqlite3_value *value,
@@ -426,31 +417,75 @@ bind_arg (sqlite3_stmt  *stmt,
 	sqlite3_bind_value (stmt, idx, value);
 }
 
+static TrackerProperty *
+get_column_property (TrackerTriplesCursor *cursor,
+                     int                   n_col)
+{
+	TrackerProperty *property;
+	const gchar *col_name;
+	int n_cols;
+
+	n_cols = sqlite3_column_count (cursor->stmt);
+	g_assert ((guint) n_cols <= cursor->n_alloc_column_properties);
+
+	if (n_col < 0 || n_col >= n_cols)
+		return NULL;
+
+	property = cursor->column_properties[n_col];
+
+	if (!property) {
+		col_name = sqlite3_column_name (cursor->stmt, n_col);
+		property = tracker_ontologies_get_property_by_uri (cursor->vtab->module->ontologies,
+		                                                   col_name);
+		cursor->column_properties[n_col] = property;
+	}
+
+	return property;
+}
+
 static gboolean
 iterate_next_stmt (TrackerTriplesCursor  *cursor,
                    const gchar          **graph,
                    TrackerRowid          *graph_id,
+                   TrackerClass         **class,
                    TrackerProperty      **property)
 {
 	TrackerRowid *id;
 
-	while (cursor->properties && !cursor->graphs) {
-		/* Iterate to next property, and redo graph list */
-		cursor->properties = g_list_remove (cursor->properties,
-		                                    cursor->properties->data);
-		cursor->graphs = g_hash_table_get_keys (cursor->query_graphs);
-	}
+	*graph_id = 0;
+	*graph = NULL;
+	*class = NULL;
+	*property = NULL;
 
-	if (!cursor->properties)
+	if (cursor->finished)
 		return FALSE;
 
-	*property = cursor->properties->data;
+	if (cursor->cur_class)
+		cursor->cur_class = cursor->cur_class->next;
+	else if (cursor->cur_property)
+		cursor->cur_property = cursor->cur_property->next;
 
-	id = cursor->graphs->data;
+	if (!cursor->cur_class && !cursor->cur_property) {
+		if (cursor->cur_graph)
+			cursor->cur_graph = cursor->cur_graph->next;
+		else
+			cursor->cur_graph = cursor->graphs;
+
+		cursor->cur_class = cursor->classes;
+		cursor->cur_property = cursor->properties;
+	}
+
+	if (!cursor->cur_graph)
+		return FALSE;
+
+	id = cursor->cur_graph->data;
 	*graph_id = *id;
 	*graph = g_hash_table_lookup (cursor->query_graphs, id);
 
-	cursor->graphs = g_list_remove (cursor->graphs, cursor->graphs->data);
+	if (cursor->cur_class)
+		*class = cursor->cur_class->data;
+	else if (cursor->cur_property)
+		*property = cursor->cur_property->data;
 
 	return TRUE;
 }
@@ -459,49 +494,70 @@ static int
 init_stmt (TrackerTriplesCursor *cursor)
 {
 	TrackerProperty *property;
+	TrackerClass *class;
 	const gchar *graph;
 	TrackerRowid graph_id;
-	GString *sql;
-	int rc;
+	int rc = SQLITE_DONE;
 
-	while (iterate_next_stmt (cursor, &graph, &graph_id, &property)) {
-		gchar *string_expr;
-
-		string_expr = convert_to_string (tracker_property_get_name (property),
-						 tracker_property_get_data_type (property));
+	while (iterate_next_stmt (cursor, &graph, &graph_id, &class, &property)) {
+		GString *sql;
 
 		sql = g_string_new (NULL);
-		g_string_append_printf (sql,
-		                        "SELECT %" G_GINT64_FORMAT ", t.ID, "
-		                        "       (SELECT ID From Resource WHERE Uri = \"%s\"), "
-		                        "       %s, "
-		                        "       %d, "
-		                        "       t.\"%s\" "
-		                        "FROM \"%s\".\"%s\" AS t "
-		                        "WHERE 1 ",
-		                        graph_id,
-		                        tracker_property_get_uri (property),
-		                        string_expr,
-		                        tracker_property_get_data_type (property),
-		                        tracker_property_get_name (property),
-		                        graph,
-		                        tracker_property_get_table_name (property));
+
+		if (class) {
+			TrackerProperty **properties;
+			guint n_properties, i;
+
+			g_string_append_printf (sql,
+			                        "SELECT %" G_GINT64_FORMAT ", ROWID ",
+			                        graph_id);
+
+
+			properties = tracker_ontologies_get_properties (cursor->vtab->module->ontologies,
+			                                                &n_properties);
+			for (i = 0; i < n_properties; i++) {
+				if (!tracker_property_get_multiple_values (properties[i]) &&
+				    tracker_property_get_domain (properties[i]) == class) {
+					g_string_append_printf (sql, ", \"%s\" ",
+					                        tracker_property_get_name (properties[i]));
+				}
+			}
+
+			g_string_append_printf (sql,
+			                        "FROM \"%s\".\"%s\" AS t ",
+			                        graph,
+			                        tracker_class_get_name (class));
+		} else if (property) {
+			if (tracker_property_get_multiple_values (property)) {
+				g_string_append_printf (sql,
+				                        "SELECT %" G_GINT64_FORMAT ", * "
+				                        "FROM \"%s\".\"%s\" AS t ",
+				                        graph_id,
+				                        graph,
+				                        tracker_property_get_table_name (property));
+			} else {
+				g_string_append_printf (sql,
+				                        "SELECT %" G_GINT64_FORMAT ", ROWID, \"%s\" "
+				                        "FROM \"%s\".\"%s\" AS t ",
+				                        graph_id,
+				                        tracker_property_get_name (property),
+				                        graph,
+				                        tracker_property_get_table_name (property));
+			}
+		}
 
 		if (cursor->match.subject) {
-			g_string_append (sql, "AND t.ID ");
+			g_string_append (sql, "WHERE t.ID ");
 			add_arg_check (sql, cursor->match.subject,
 			               !!(cursor->match.idxFlags & IDX_MATCH_SUBJECT_NEG),
 			               "@s");
 		}
 
 		rc = sqlite3_prepare_v2 (cursor->vtab->module->db,
-					 sql->str, -1, &cursor->stmt, 0);
+		                         sql->str, -1, &cursor->stmt, 0);
 		g_string_free (sql, TRUE);
-		g_free (string_expr);
 
 		if (rc == SQLITE_OK) {
-			if (cursor->match.graph)
-				bind_arg (cursor->stmt, cursor->match.graph, "@g");
 			if (cursor->match.subject)
 				bind_arg (cursor->stmt, cursor->match.subject, "@s");
 
@@ -509,12 +565,29 @@ init_stmt (TrackerTriplesCursor *cursor)
 		}
 
 		if (rc != SQLITE_DONE)
-			return rc;
+			break;
 
 		g_clear_pointer (&cursor->stmt, sqlite3_finalize);
 	}
 
-	return SQLITE_DONE;
+	if (rc == SQLITE_ROW) {
+		int columns;
+
+		columns = sqlite3_column_count (cursor->stmt);
+
+		if ((guint) columns > cursor->n_alloc_column_properties) {
+			g_free (cursor->column_properties);
+			cursor->column_properties =
+				g_new0 (TrackerProperty*, columns);
+			cursor->n_alloc_column_properties = columns;
+		} else {
+			bzero (cursor->column_properties,
+			       sizeof (TrackerProperty*) *
+			       cursor->n_alloc_column_properties);
+		}
+	}
+
+	return rc;
 }
 
 static int
@@ -549,8 +622,14 @@ triples_filter (sqlite3_vtab_cursor  *vtab_cursor,
 	if ((rc = collect_graphs (cursor)) != SQLITE_DONE)
 		return rc;
 
-	collect_properties (cursor);
+	collect_tables (cursor);
+
+	cursor->cur_graph = NULL;
+	cursor->cur_class = NULL;
+	cursor->cur_property = NULL;
+
 	rc = init_stmt (cursor);
+	cursor->column = FIRST_PROPERTY_COLUMN;
 
 	if (rc == SQLITE_DONE)
 		cursor->finished = TRUE;
@@ -565,21 +644,43 @@ static int
 triples_next (sqlite3_vtab_cursor *vtab_cursor)
 {
 	TrackerTriplesCursor *cursor = (TrackerTriplesCursor *) vtab_cursor;
-	int rc;
+	int rc, column_count;
+
+	cursor->rowid++;
+	column_count = sqlite3_column_count (cursor->stmt);
+
+	while (cursor->column < column_count) {
+		cursor->column++;
+
+		if ((cursor->match.idxFlags & IDX_MATCH_PREDICATE_NEG) != 0) {
+			TrackerProperty *property;
+
+			/* Single valued properties skip the "predicate != ..." here */
+			property = get_column_property (cursor, cursor->column);
+
+			if (property &&
+			    sqlite3_value_int64 (cursor->match.predicate) ==
+			    tracker_property_get_id (property))
+				cursor->column++;
+		}
+
+		if (sqlite3_column_type (cursor->stmt, cursor->column) != SQLITE_NULL)
+			break;
+	}
+
+	if (cursor->column < column_count)
+		return SQLITE_OK;
 
 	rc = sqlite3_step (cursor->stmt);
+	cursor->column = FIRST_PROPERTY_COLUMN;
 
 	if (rc == SQLITE_DONE) {
 		g_clear_pointer (&cursor->stmt, sqlite3_finalize);
 		rc = init_stmt (cursor);
 	}
 
-	if (rc == SQLITE_ROW) {
-		cursor->rowid++;
-	} else {
+	if (rc != SQLITE_ROW)
 		cursor->finished = TRUE;
-	}
-
 	if (rc != SQLITE_ROW && rc != SQLITE_DONE)
 		return rc;
 
@@ -601,12 +702,34 @@ triples_column (sqlite3_vtab_cursor *vtab_cursor,
 {
 	TrackerTriplesCursor *cursor = (TrackerTriplesCursor *) vtab_cursor;
 	sqlite3_value *value;
+	TrackerProperty *property;
 
-	if (n_col == COL_ROWID) {
-		sqlite3_result_int64 (context, cursor->rowid);
-	} else {
-		value = sqlite3_column_value (cursor->stmt, n_col - 1);
+	switch (n_col) {
+	case COL_GRAPH:
+		value = sqlite3_column_value (cursor->stmt, 0);
 		sqlite3_result_value (context, value);
+		break;
+	case COL_SUBJECT:
+		value = sqlite3_column_value (cursor->stmt, 1);
+		sqlite3_result_value (context, value);
+		break;
+	case COL_OBJECT:
+		value = sqlite3_column_value (cursor->stmt, cursor->column);
+		sqlite3_result_value (context, value);
+		break;
+	case COL_PREDICATE:
+	case COL_OBJECT_TYPE:
+		property = get_column_property (cursor, cursor->column);
+		if (!property) {
+			sqlite3_result_error_code (context, SQLITE_CORRUPT);
+			break;
+		}
+
+		if (n_col == COL_PREDICATE)
+			sqlite3_result_int64 (context, tracker_property_get_id (property));
+		else
+			sqlite3_result_int64 (context, tracker_property_get_data_type (property));
+		break;
 	}
 
 	return SQLITE_OK;
