@@ -142,7 +142,7 @@ def create_sandbox(store_location, prefix=None, use_session_dirs=False,
 
 def config_set(sandbox, content_locations_recursive=None,
                content_locations_single=None, applications=False):
-    dconfclient = dconf.DConfClient(sandbox)
+    dconfclient = sandbox.get_dconf_client()
 
     if content_locations_recursive:
         log.debug("Using content locations: %s" %
@@ -255,8 +255,8 @@ def init_logging(debug_sandbox, debug_dbus):
         root.setLevel(logging.DEBUG)
         root.addHandler(sandbox_log_handler)
     else:
-        dbus_stderr = logging.getLogger('trackertestutils.dbusdaemon.stderr')
-        dbus_stdout = logging.getLogger('trackertestutils.dbusdaemon.stdout')
+        dbus_stderr = logging.getLogger('sandbox-session-bus.stderr')
+        dbus_stdout = logging.getLogger('sandbox-session-bus.stdout')
 
         dbus_handler = logging.StreamHandler(stream=sys.stderr)
         dbus_handler.setFormatter(logging.Formatter(DBUS_FORMAT))
